/***********************************************************************/
/*                                                                     */
/*               Output Analysis Program for ETMSP (OAP)               */
/*               =======================================               */
/*                                                                     */
/*        copyright @(1993) Electric Power Research Institute          */
/*                                                                     */
/* EPRI reserves all rights in the software and associated             */
/* documentation as delivered. The software and its associated         */
/* documention or any portion thereof may not be reproduced in any form*/
/* whatsoever without the consent of EPRI, except as explicity provided*/
/* for by license. Licensee shall submit a detailed description to EPRI*/
/* of any changes or modifications to the software within two (2)      */
/* months after such changes or modifications are made. At EPRI's sole */
/* discretion, the software, as changed or modified, shall be given a  */
/* new designation sufficiently different from its then current        */
/* designation as to prevent mistake,confusion, or deception as between*/
/* the current software as changed or modified.                        */
/*                                                                     */
/* A license under EPRI's rights in the software can be obtained       */
/* directly from EPRI.                                                 */
/*                                                                     */
/* This program and supporting materials may be independently available*/
/* from Electric Power Software centre (EPSC) for an appropriate       */
/* distribution fee. The EPSC is a service of                          */
/*                                                                     */
/*                    Power Computing Company                          */
/*                    1930 Hi Line                                     */
/*                    Dallas, Tx 75207                                 */
/*                                                                     */
/* NEITHER EPRI, ANY MEMBER OF EPRI, NOR ANY PERSON OR  ORGANIZATION   */
/* ACTING ON BEHAVE OF THEM:                                           */
/*                                                                     */
/* 1. MAKE ANY WARRANTY OR REPRESENTATION WHATSOEVER, EXPRESS OR       */
/*    IMPLIED, INCLUDING ANY WARRANTY OF MERCHANTABILITY OR FITNESS OR */
/*    ANY PURPOSE WITH RESPECT TO THE SOFTWARE OR ITS ASSOCIATED       */
/*    DOCUMENTATION; OR                                                */
/*                                                                     */
/* 2. ASSUMES ANY LIABILITY WHATSOEVER WITH RESPECT TO ANY USE OF THE  */
/*    SOFTWARE OR ITS ASSOCIATED DOCUMENTATION OR ANY PORTION THEREOF  */
/*    OR WITH RESPECT TO ANY DAMAGES WHICH MAY RESULT FROM SUCH USE.   */
/*                                                                     */
/* THIS NOTICE MAY NOT BE REMOVED FROM THE SOFTWARE BY ANY USER        */
/* THEREOF.                                                            */
/*                                                                     */
/***********************************************************************/

/***********************************************************
 *
 * File:          ps3d.c 
 * Description:   save the plot to a PostScript format file for 
 *                printing on a PostScript printer
 *
 ***********************************************************/

/*
 *    Include File
 */

#include "X3d_PS.h"


/*
 *    Function Declarations
 */

void rotate_label();
void tick_ps();

extern int textlen();
extern get_time();
extern vector new_p_coordinate();
extern float surface_check();
extern wp_scale();
extern matrix_op_mul();
extern indexx();

/*
 *    Global Variables and Type Definition
 */

matrix trans_ps;
FILE *fp;
int Pxmin, Pymin, Pxmax, Pymax;
int PPxmin, PPymin, PPxmax, PPymax;
float PX_DIM, PY_DIM, PRightMargin, PTopMargin;
int TitleFontSize,SubTitleFontSize,TextFontSize,AnnotationFontSize;
typedef struct {
  float x;
  float y;
} Point_2d;
char fn3d[50];

/***********************************************************
 * Page Setup
 *
 * PRightMargin: default right margin in inches
 * PTopMargin:   default top margin in inches
 * PX_DIM: x dimension of plotting area in inches
 * PY_DIM: y dimension of plotting area in inches
 * Pxmin:  minimum X coordinate of the frame in inches
 * Pxmax:  maximun X coordinate of the frame in inches
 * Pymin:  minimum Y coordinate of the frame in inches
 * Pymax:  maximun Y coordinate of the frame in inches
 * PPxmin: minimum X coordinate of the plot in inches
 * PPxmax: maximun X coordinate of the plot in inches
 * PPymin: minimum Y coordinate of the plot in inches
 * PPymax: maximun Y coordinate of the plot in inches
 */

void ps_3d_page_setup_()
{
   extern int AnnotationFontSize,TextFontSize,
	TitleFontSize;
   extern float PX_DIM, PY_DIM, PRightMargin, PTopMargin;
   extern int Pxmin, Pymin, Pxmax, Pymax;
   extern int PPxmin, PPymin, PPxmax, PPymax;

   /* default page setting */
   PX_DIM = 7.;         
   PY_DIM = 7.; 
   PRightMargin = 0.75;
   PTopMargin = 2.8;
   TitleFontSize = 12;
   SubTitleFontSize = 8;
   TextFontSize =9;
   AnnotationFontSize = 9;

   /* coordinates of four corners in plotting area */ 
   Pxmax = (int)((8.5 - PRightMargin)*72);
   Pxmin = (int)(((float)Pxmax/72. - PX_DIM)*72);
   Pymax = (int)((11. - PTopMargin)*72);
   Pymin = (int)(((float)Pymax/72. - PY_DIM)*72); 
   PPxmin=Pxmin+72;
   PPxmax=Pxmax-72;
   PPymin=Pymin+72;
   PPymax=Pymax-72;
}

/***********************************************************
 * open PostScript file
 * -- write prologue part to PostScript file
 * -- draw logo, frame and page title
 */

void ps_3d_begin_(filename, lenfn, imn)
char *filename;    /* PostScript file to be saved under */
int  *lenfn,       /* string length of filename */
     *imn;         /* if imn=0, open file; otherwise, suppose
                      the file's been opened */ 
{
   extern int AnnotationFontSize,TextFontSize,
        TitleFontSize;
   extern int Pxmin, Pymin, Pxmax, Pymax;
   char newtime[80];
    int i, j;
#ifdef VAX
   str_desc *filenm_dsc;
#endif

   /* for VAX, must typecase filename arg to a descriptor struc pointer */
#ifdef VAX
   filenm_dsc = (str_desc *) filename;
   strncpy (fn3d, filenm_dsc ->dsc$a_pointer, *lenfn);
#else
   strncpy (fn3d, filename, *lenfn);
#endif

   fn3d[*lenfn] = '\0';

   /* get local time */
     get_time(newtime);
  
   /* open PostScript file */
   if (!*imn) {
      if ((fp = fopen(fn3d,"w")) == NULL) {
         fprintf(stderr,"cat: can't open %s\n",*filename);
         exit(1);
     }
   }

   /* header */
   fprintf(fp,"%%!PS-Adobe-2.0 EPSF-1.2\n");
   fprintf(fp,"%%%%BoundingBox: 24 24 588 768\n");
   fprintf(fp,"%%%%OAP postscript outputfile\n");
   fprintf(fp,"%%%%Creator: Ontario Hydro\n");
   fprintf(fp,"%%%%Creation Date: %s\n",newtime);

   /* process second page */
   if (*imn) goto Pg_begin;   

   /* if it's the first page, define procedures */
   fprintf(fp,"/inch {72 mul} def\n");
   fprintf(fp,"/TitleFontSize %4d def\n",TitleFontSize);      
   fprintf(fp,"/SubTitleFontSize %4d def\n",SubTitleFontSize);  
   fprintf(fp,"/AnnotationFontSize %4d def\n",AnnotationFontSize);  
   fprintf(fp,"/TextFontSize %4d def\n",TextFontSize);       
   fprintf(fp,"/FS { findfont exch scalefont setfont } bind def\n");
   fprintf(fp,"/M {moveto} def\n");
   fprintf(fp,"/R {rmoveto} def\n");
   fprintf(fp,"/Lt {lineto} def\n");
   fprintf(fp,"/L {rlineto} def\n");
   fprintf(fp,"/S {stroke} def\n");
   fprintf(fp,"/Sd {setdash} def\n");
   fprintf(fp,"/Sg {setgray} def\n");
   fprintf(fp,"/Sl {setlinewidth} def\n");
   fprintf(fp,"/Gs {gsave} def\n");
   fprintf(fp,"/Gr {grestore} def\n");
   fprintf(fp,"/MoveLeft {dup stringwidth pop neg 0 R} def\n");
   fprintf(fp,"/MoveCenter {dup stringwidth pop -2 div 0 R} def\n");
   fprintf(fp,"/MoveDown {neg 0 exch R} def\n"); 
   fprintf(fp,"/MoveUp {0 exch R} def\n");
   fprintf(fp,"/Coord {M currentpoint translate} def\n");
   fprintf(fp,"/Tick {M Lt S} def\n");
   fprintf(fp,"/Sf {newpath M Lt Lt Lt closepath Gs 1 Sg fill Gr S} def\n");
   fprintf(fp,"/P4 {newpath M Lt Lt Lt closepath S} def\n");
   fprintf(fp,"/Pop4 {4 {pop} repeat}def\n");
   fprintf(fp,"/Hydro {0 8 M 0 32 32 32 8 arcto Pop4\n");
   fprintf(fp,"    32 32 32 0 8 arcto Pop4\n");
   fprintf(fp,"    32 0 0 0 8 arcto Pop4\n");
   fprintf(fp,"    0 0 0 32 8 arcto Pop4}def\n");

   Pg_begin:                  /* start new page */
   fprintf(fp,"save\n");
   fprintf(fp,"1 1 scale\n");
#ifdef _LOGO_OH
   /* draw hydro logo */
   fprintf(fp,"Gs 1.2 inch 9.75  inch Coord\n");
   fprintf(fp,"45 rotate\n");
   fprintf(fp,"Hydro 0.5 Sg fill Hydro 0 Sg 6 Sl\n");
   fprintf(fp,"11 -20 R 0 24 L 11 0 R 0 -24 L\n");
   fprintf(fp,"S Gr\n");
   fprintf(fp,"12 /Helvetica-Bold FS\n");
   fprintf(fp,"1.7 inch  9.95 inch M\n");
   fprintf(fp,"(ontario hydro) show\n");
#endif
#ifdef _LOGO_EPRI   
   /* EPRI logo, if needed */
   fprintf(fp,"28 /Helvetica-Bold FS\n");
   fprintf(fp,"0.8 inch 8 inch M\n");
   fprintf(fp,"28 MoveDown\n");
   fprintf(fp,"(EPRI) show\n");
   fprintf(fp,"10 /Times-Bold  FS\n");
   fprintf(fp,"0.8 inch 8 inch M\n");
   fprintf(fp,"3 -40 R\n");
   fprintf(fp,"(Leadership in Science and Technology) show\n");
#endif  
   /* draw frame and page title */
   fprintf(fp,"1. Sl [] 0 Sd\n");
   fprintf(fp,"%6d %6d M\n",Pxmin,Pymin);
   fprintf(fp," 0 %6d L\n",Pymax-Pymin);
   fprintf(fp," %6d 0 L\n",Pxmax-Pxmin);
   fprintf(fp," 0 %6d L\n",Pymin-Pymax);
   fprintf(fp,"closepath S\n");

   fprintf(fp,"TitleFontSize /Times-Bold FS\n");
    i = 6;

    while(i && strspn(ptit_.ptitle[--i], " ")); i++;
    fprintf(fp,"%d %d M\n", Pxmin, Pymax+47+i*7);
    fprintf(fp,"(%.12s) show\n",ptit_.caseid[0]);
    for(j=0; j<i; j++) {
        fprintf(fp,"%d %d M\n", Pxmin, Pymax+33-j*14+i*7);
        fprintf(fp,"(%.80s) show\n",ptit_.ptitle[j]);
    }
    fprintf(fp,"%d %d M\n", (Pxmin+Pxmax)/2, Pymin-20);
    /* ptit_.ftitle[lens_(ptit_.ftitle)] = '\0'; */
    ptit_.ftitle[textlen(ptit_.ftitle)] = '\0';
    fprintf(fp,"(%s) MoveCenter show\n",ptit_.ftitle);

    fprintf(fp,"SubTitleFontSize /Courier FS\n");
    fprintf(fp,"%d %d M\n", Pxmax, Pymax+47+i*7);
    fprintf(fp,"(%s) MoveLeft show\n",newtime);
}

/***********************************************************
 * transformation matrix for coordinates on the page 
 */
void transfor_ps_()
{
   extern matrix  Mr,trans_ps;
   extern int PPxmin, PPymin, PPxmax, PPymax;
   matrix p_scal;

   /* get scaling matrix for page coordinate */
   wp_scale(Mr,p_scal,PPxmin,PPxmax,PPymin,PPymax);

   /* transformation matrix for page coordinate */
   matrix_op_mul(trans_ps,Mr,p_scal);
}

/***********************************************************
 * wirte label along rotated coordinate
 */
void rotate_label(vst,ved)
vector vst, /* coordinates of the starting point of the axis */ 
       ved; /* coordinates of the end point of the axis */
{
   extern matrix trans_ps;
   float r_angle,dx,dy;
   vector new_vst, new_ved;

   new_vst=new_p_coordinate(vst,trans_ps);
   new_ved=new_p_coordinate(ved,trans_ps);
   fprintf(fp,"%8.2f %8.2f Coord\n",
	(new_vst.x+new_ved.x)/2.,(new_vst.y+new_ved.y)/2. );
   if (new_vst.y < new_ved.y) {
    	dy = (new_ved.y-new_vst.y)/2.;
    	dx = (new_ved.x-new_vst.x)/2.; }
   else {
    	dy = (new_vst.y-new_ved.y)/2.;
    	dx = (new_vst.x-new_ved.x)/2.; }
   r_angle = atan(dy/dx)*RAD_D;
   if (r_angle>=-95. && r_angle<=-85.) r_angle=180.+r_angle;
   fprintf(fp,"%8.2f rotate\n",r_angle); 
} 

/***********************************************************
 * draw ticks and labels for X,Y,Z axes
 */
void tick_ps(id, vst, ved, rmin, rmax, ratio, ddx, ddy, ddz,dl)
vector vst, ved;
int id;
float rmin,rmax,ratio,ddx,ddy,ddz,dl;
{
   extern matrix trans_ps;
   extern char xlab[80], ylab[80], zlab[80];
   vector tem1,tem2,newt1,newt2;
   float dd,dc,t_label;
   char text[30];
   int i;

   dc=(rmax-rmin)/5.;
   switch (id)
   {
    case 0:  /* x axis ticks and annotations */
      dd=(ved.x-vst.x)/5.;
      fprintf(fp,"TextFontSize /Times-Roman FS\n");
      fprintf(fp,"1. Sl [] 0 Sd\n");
      for (i=0; i<=5; i++) {
        tem1.x = tem2.x = vst.x +i*dd;
        tem1.y = vst.y-ddy; tem2.y = vst.y +ddy;
        tem1.z = vst.z-ddz; tem2.z = vst.z +ddz;;
        newt1 = new_p_coordinate(tem1,trans_ps);
        newt2 = new_p_coordinate(tem2,trans_ps);
   	fprintf(fp,"%8.2f %8.2f ",newt1.x,newt1.y);
   	fprintf(fp,"%8.2f %8.2f Tick\n",newt2.x,newt2.y);
        t_label=rmin+i*dc;
        tem1.x = vst.x+i*dd;
        tem1.y = vst.y + ratio*ddy; tem1.z = vst.z + ratio*ddz;
        newt1 = new_p_coordinate(tem1,trans_ps);
        sprintf(text,"%4.2f",t_label);
   	fprintf(fp,"%8.2f  %8.2f M ",newt1.x,newt1.y);
   	fprintf(fp,"(%s) show\n",text);
      }

      /* x label */
      dd=ratio/fabs(ratio);
      fprintf(fp,"Gs TextFontSize /Times-Bold FS\n");
      tem1.x=vst.x; tem1.y = vst.y+dl*dd*ddy; tem1.z=vst.z+dl*dd*ddz;
      tem2.x=ved.x; tem2.y = ved.y+dl*dd*ddy; tem2.z=ved.z+dl*dd*ddz;
      rotate_label(tem1,tem2);
      fprintf(fp,"(%s) MoveCenter show Gr\n",xlab);
      break;

    case 1: /* y axis ticks and annotations */
      dd=(ved.y-vst.y)/5.;
      fprintf(fp,"TextFontSize /Times-Roman FS\n");
      fprintf(fp,"1. Sl [] 0 Sd\n");
      for (i=0; i<=5; i++) {
        tem1.y = tem2.y = vst.y+i*dd;
        tem1.x = vst.x - ddx; tem2.x = vst.x + ddx;
        tem1.z = vst.z -ddz; tem2.z  = vst.z + ddz;
        newt1 = new_p_coordinate(tem1,trans_ps);
        newt2 = new_p_coordinate(tem2,trans_ps);
        fprintf(fp,"%8.2f %8.2f ",newt1.x,newt1.y);
        fprintf(fp,"%8.2f %8.2f Tick\n",newt2.x,newt2.y);
        t_label=rmin+i*dc;
        tem1.y = vst.y+i*dd;
        tem1.x = vst.x + ratio*ddx; tem1.z = vst.z+ratio*ddz;
        newt1 = new_p_coordinate(tem1,trans_ps);
        sprintf(text,"%4.2f",t_label);
        fprintf(fp,"%8.2f  %8.2f M ",newt1.x,newt1.y);
        fprintf(fp,"(%s) show\n",text);
      }

      /* y label */
      dd=ratio/fabs(ratio);
      fprintf(fp,"Gs TextFontSize /Times-Bold FS\n");
      tem1.y=vst.y; tem1.x = vst.x+dl*dd*ddx; tem1.z=vst.z+dl*dd*ddz;
      tem2.y=ved.y; tem2.x = ved.x+dl*dd*ddx; tem2.z=ved.z+dl*dd*ddz;
      rotate_label(tem1,tem2);
      fprintf(fp,"(%s) MoveCenter show Gr\n",ylab);
      break;

    case 2: /* z axis ticks and annotations */
      fprintf(fp,"TextFontSize /Times-Roman FS\n");
      fprintf(fp,"1. Sl [] 0 Sd\n");
      dd=(ved.z-vst.z)/5.;
      for (i=0; i<=5; i++) {
        tem1.z = tem2.z = vst.z+i*dd;
        tem1.y = vst.y -ddy; tem2.y = vst.y +ddy;
        tem1.x = vst.x-ddx; tem2.x = vst.x+ddx;
        newt1 = new_p_coordinate(tem1,trans_ps);
        newt2 = new_p_coordinate(tem2,trans_ps);
        fprintf(fp,"%8.2f %8.2f ",newt1.x,newt1.y);
        fprintf(fp,"%8.2f %8.2f Tick\n",newt2.x,newt2.y);
        t_label=rmin+i*dc;
        tem1.z = vst.z+i*dd;
        tem1.x = vst.x +ratio*ddx;
        tem1.y = vst.y +ratio*ddy;
        newt1 = new_p_coordinate(tem1,trans_ps);
        sprintf(text,"%4.2f",t_label);
        fprintf(fp,"%8.2f  %8.2f M ",newt1.x,newt1.y);
        fprintf(fp,"(%s) show\n",text);
      }

      /* z label */
      dd=ratio/fabs(ratio);
      fprintf(fp,"Gs TextFontSize /Times-Bold FS\n");
      tem1.z=vst.z; tem1.x = vst.x+dl*dd*ddx; tem1.y=vst.y+dl*dd*ddy;
      tem2.z=ved.z; tem2.x = ved.x+dl*dd*ddx; tem2.y=ved.y+dl*dd*ddy;
      rotate_label(tem1,tem2);
      fprintf(fp,"(%s) MoveCenter show Gr\n",zlab);
      break;

    default: break;
  }
}

/***********************************************************
 * decide which axes ticks and labels should be put on
 */
void axes_ps_()
{
   vector new_oo,new_xo,new_yo,new_zo,new_xoy,new_yoz,new_zox,new_xyz;
   float test,t[4],rl,rr;
   int i, id,xy_tick_index;

   /* get page coordinates of vector oo, xo, yo, zo,
      xoy, yoz, zox, xyz */ 
   new_oo=new_p_coordinate(oo,trans_ps);
   new_xo=new_p_coordinate(xo,trans_ps);
   new_yo=new_p_coordinate(yo,trans_ps);
   new_zo=new_p_coordinate(zo,trans_ps);
   new_xoy=new_p_coordinate(xoy,trans_ps);
   new_yoz=new_p_coordinate(yoz,trans_ps);
   new_zox=new_p_coordinate(zox,trans_ps);
   new_xyz=new_p_coordinate(xyz,trans_ps);

   /* draw cube */
   fprintf(fp,"1. Sl [4 3] 0 Sd\n");
   fprintf(fp,"%8.2f %8.2f M\n",new_xyz.x,new_xyz.y);
   fprintf(fp,"%8.2f %8.2f Lt\n",new_xoy.x,new_xoy.y);
   fprintf(fp,"%8.2f %8.2f Lt\n",new_yo.x,new_yo.y);
   fprintf(fp,"%8.2f %8.2f Lt\n",new_yoz.x,new_yoz.y);
   fprintf(fp,"%8.2f %8.2f Lt\n",new_zo.x,new_zo.y);
   fprintf(fp,"%8.2f %8.2f Lt\n",new_zox.x,new_zox.y);
   fprintf(fp,"%8.2f %8.2f Lt\n",new_xyz.x,new_xyz.y);
   fprintf(fp,"%8.2f %8.2f Lt\n",new_yoz.x,new_yoz.y);
   fprintf(fp,"%8.2f %8.2f M\n",new_xoy.x,new_xoy.y);
   fprintf(fp,"%8.2f %8.2f Lt\n",new_xo.x,new_xo.y);
   fprintf(fp,"%8.2f %8.2f Lt\n",new_oo.x,new_oo.y);
   fprintf(fp,"%8.2f %8.2f Lt\n",new_yo.x,new_yo.y);
   fprintf(fp,"%8.2f %8.2f M\n",new_xo.x,new_xo.y);
   fprintf(fp,"%8.2f %8.2f Lt\n",new_zox.x,new_zox.y);
   fprintf(fp,"%8.2f %8.2f M\n",new_oo.x,new_oo.y);
   fprintf(fp,"%8.2f %8.2f Lt S\n",new_zo.x,new_zo.y);

   /* draw ticks, annotations ans labels */
   if(tickflag==8) {
        t[0]=new_oo.x+new_zo.x;
        t[1]=new_xoy.x+new_xyz.x;
        t[2]=new_xo.x+new_zox.x;
        t[3]=new_yo.x+new_yoz.x;
        test=100000.;
        for (i=0; i<4; i++){
           if (t[i]<test) {
                test=t[i];
                id=i; }
        }
        t[0]=0.02*(xmax-xmin);
        t[1]=0.02*(ymax-ymin);
        if (id==0) {
          tick_ps(2, oo, zo, zmin, zmax, -6.,t[0], t[1], 0.,10.);
          if (new_zo.y>new_oo.y)
            if (new_yo.y>new_xo.y)
                xy_tick_index=0;
            else
                xy_tick_index=2;
          else
            if (new_yo.y<new_xo.y)
                xy_tick_index=1;
            else
                xy_tick_index=3;
        }
        if (id==1) {
          tick_ps(2, xoy, xyz, zmin, zmax,6.,t[0], t[1], 0.,10.);
          if (new_xyz.y>new_xoy.y)
            if (new_yo.y<new_xo.y)
                xy_tick_index=2;
            else
                xy_tick_index=0;
          else
            if (new_yo.y>new_xo.y)
                xy_tick_index=3;
            else
                xy_tick_index=1;
        }
        if (id==2) {
          tick_ps(2, xo, zox, zmin, zmax, 6.,t[0], -t[1], 0.,10.);
          if (new_zox.y>new_xo.y)
            if (new_xoy.y < new_oo.y)
                xy_tick_index=4;
            else
                xy_tick_index=6;
          else
            if (new_xoy.y > new_oo.y)
                xy_tick_index=5;
            else
                xy_tick_index=7;
        }
        if (id==3) {
          tick_ps(2, yo, yoz, zmin, zmax, 6.,-t[0], t[1] , 0.,10.);
          if (new_yoz.y>new_yo.y)
            if (new_xoy.y > new_oo.y)
                xy_tick_index=6;
            else
                xy_tick_index=4;
          else
            if (new_xoy.y < new_oo.y)
                xy_tick_index=7;
            else
                xy_tick_index=5;
        }
        switch(xy_tick_index) {
         case 0:
	   if ((new_oo.x+new_xo.x)<(new_xoy.x+new_xo.x)) {rl=6.; rr=9.;}
	   else   {rl=9.; rr=6.;}
           tick_ps(0, oo, xo, xmin, xmax, -rr, 0., .02*(ymax-ymin), 0.,18.);
           tick_ps(1, xo, xoy, ymin, ymax, rl,.02*(xmax-xmin), 0.,0.,18.);
           break;
         case 1:
           if ((new_yoz.x+new_xyz.x)<(new_yoz.x+new_zo.x)) {rl=6.; rr=9.;}
           else   {rl=9.; rr=6.;}
           tick_ps(0, yoz, xyz,xmin, xmax, rr,0.,.02*(ymax-ymin), 0.,18.);
           tick_ps(1, zo, yoz,  ymin, ymax,-rl,.02*(xmax-xmin), 0.,0.,18.);
           break;
         case 2:
           if ((new_yo.x+new_xoy.x)<(new_yo.x+new_oo.x)) {rl=6.; rr=9.;}
           else   {rl=9.; rr=6.;}
           tick_ps(0, yo, xoy, xmin, xmax, rr, 0., .02*(ymax-ymin), 0.,18.);
           tick_ps(1, oo, yo, ymin, ymax, -rl,.02*(xmax-xmin), 0., 0.,18.);
           break;
         case 3:
           if ((new_zo.x+new_zox.x)<(new_xyz.x+new_zox.x)) {rl=6.; rr=9.;}
           else   {rl=9.; rr=6.;}
           tick_ps(0, zo, zox,xmin, xmax, -rr, 0., .02*(ymax-ymin), 0.,18.);
           tick_ps(1, zox, xyz, ymin, ymax, rl,.02*(xmax-xmin), 0., 0.,18.);
           break;
         case 4:
           if ((new_yo.x+new_xoy.x)<(new_xoy.x+new_xo.x)) {rl=6.; rr=9.;}
           else   {rl=9.; rr=6.;}
           tick_ps(0, yo, xoy, xmin, xmax, rr, 0., .02*(ymax-ymin), 0.,18.);
           tick_ps(1, xo, xoy, ymin, ymax, rl,.02*(xmax-xmin), 0.,0.,18.);
           break;
         case 5:
           if ((new_zo.x+new_zox.x)<(new_yoz.x+new_zo.x)) {rl=6.; rr=9.;}
           else   {rl=9.; rr=6.;}
           tick_ps(0, zo, zox,  xmin, xmax, -rr, 0., .02*(ymax-ymin), 0.,18.);
           tick_ps(1, zo, yoz,  ymin, ymax,-rl,.02*(xmax-xmin), 0.,0.,18.);
           break;
         case 6:
           if ((new_oo.x+new_xo.x)<(new_yo.x+new_oo.x)) {rl=6.; rr=9.;}
           else   {rl=9.; rr=6.;}
           tick_ps(0, oo, xo, xmin, xmax, -rr, 0., .02*(ymax-ymin), 0.,18.);
           tick_ps(1, oo, yo, ymin, ymax, -rl,.02*(xmax-xmin), 0., 0.,18.);
           break;
         case 7:
           if ((new_yoz.x+new_xyz.x)<(new_zox.x+new_xyz.x)) {rl=6.; rr=9.;}
           else   {rl=9.; rr=6.;}
           tick_ps(0, yoz, xyz, xmin, xmax, rr, 0., .02*(ymax-ymin), 0.,18.);
           tick_ps(1, zox, xyz, ymin, ymax, rl,.02*(xmax-xmin), 0., 0.,18.);
           break;
        }
   }

   /* draw ticks, annotations ans labels  for XY plot  */
   if(tickflag==5){
           tick_ps(0, oo, xo, xmin, xmax, -3., 0., .02*(ymax-ymin), 0.,9.);
           tick_ps(1, oo, yo, ymin, ymax, -6.,.02*(xmax-xmin), 0., 0.,9.);
   }

   /* draw ticks, annotations ans labels  for YZ plot  */
   if(tickflag==6){
           tick_ps(1, oo, yo, ymin, ymax, -3.,0., 0.,.02*(zmax-zmin),9.);
           tick_ps(2, oo, zo, zmin, zmax, -6.,0.,.02*(ymax-ymin), 0.,9.);
   }

   /* draw ticks, annotations ans labels  for ZX plot  */
   if(tickflag==7){
           tick_ps(0, oo, xo, xmin, xmax, -3., 0.,0., .02*(zmax-zmin),9.);
           tick_ps(2, oo, zo, zmin, zmax, -6.,.02*(xmax-xmin),0., 0.,9.);
   }
}

/***********************************************************
 * draw curve in space, plot relay margins
 */
void plot3d_ps_(idx1,idx2,xtem,ytem,ztem,xaray,yaray,
       maxpt,max_p,xs, ys, xd, yd, iflg, ics, icc)
float xtem[], ytem[], ztem[];
float xaray[],yaray[];
float xs[], ys[], xd[], yd[];
int *idx1,*idx2;
int *maxpt,*max_p,*iflg;
int *icc, *ics;
{
  extern matrix trans_ps; 
  vector point, new_point, new_point1, new_point2, new_point3;
  int i,j,id,ip,ix, iyz, is,mx1,ny1,mx2,ny2;
  int icy,iz,order[8001];
  int mtemp;
  float test;
  float zid[8000];
  Point_2d xpoint[8000];

  if (*iflg==2) {
  /* plot line bindings in 3d space */
   for (i=0; i<*max_p; i++)
   {
    point.x=xtem[0];point.y=xs[i];point.z=ys[i];
    new_point1 = new_p_coordinate(point,trans_ps);
    point.x=xtem[*maxpt-1];point.y=xs[i];point.z=ys[i];
    new_point2 = new_p_coordinate(point,trans_ps);
    point.x=xtem[*maxpt-1];point.y=xd[i];point.z=yd[i];
    new_point3 = new_p_coordinate(point,trans_ps);
    point.x=xtem[i];point.y=xd[i];point.z=yd[i];
    new_point = new_p_coordinate(point,trans_ps);
    fprintf(fp,"%8.2f %8.2f %8.2f %8.2f\n %8.2f %8.2f %8.2f %8.2f P4",
	new_point1.x,new_point1.y,new_point2.x,new_point2.y,
	new_point3.x,new_point3.y,new_point.x,new_point.y);
   }
 
  /* plot relay margins */

  /* plot two end surfaces */
   iz=0;
   for(icy=0; icy<*icc; icy++)
   {
      for (i=0;i<2;i++){
   	id = i*(*maxpt-1);
     	point.x=xtem[id]; point.y=xaray[ics[icy]];
     	point.z=yaray[ics[icy]];
     	new_point = new_p_coordinate(point,trans_ps);
     	fprintf(fp," %8.2f %8.2f M\n",new_point.x,new_point.y);
     	for (j=1 ; j<maxpt[1]; j++) {
      		point.x=xtem[id]; point.y=xaray[ics[icy]+j];
      		point.z=yaray[ics[icy]+j];
      		new_point = new_p_coordinate(point,trans_ps);
      		fprintf(fp," %8.2f %8.2f Lt\n",new_point.x,new_point.y);
     	}
    	fprintf(fp,"S\n");
      }

   /* plot side surfaces */
      ix =10;
      iyz = 25;
      is = ix * iyz;
      for(i=0; i<is; i++){
     	mx1=(int)((int)(i%ix)*(*maxpt-1.)/ix);
     	ny1=(int)(floor((float)(i/ix))*(maxpt[1]-1.)/iyz);
     	if ((int)(i%ix)==(ix-1))
       	  mx2= *maxpt-1;
     	else
       	  mx2=(int)((int)(i%ix+1)*(*maxpt-1.)/ix);
        ny2=(int)(floor((float)(i/ix+1.))*(maxpt[1]-1.)/iyz);
    	point.x=xtem[mx1];point.y=xaray[ics[icy]+ny1];
    	point.z=yaray[ics[icy]+ny1];
    	new_point1= new_p_coordinate(point,trans_ps);
    	xpoint[iz*5].x = new_point1.x;
    	xpoint[iz*5].y = new_point1.y;

    	point.x=xtem[mx2];point.y=xaray[ics[icy]+ny1];
    	point.z=yaray[ics[icy]+ny1];
    	new_point2= new_p_coordinate(point,trans_ps);
    	xpoint[iz*5+1].x = new_point2.x;
    	xpoint[iz*5+1].y = new_point2.y;

    	point.x=xtem[mx2];point.y=xaray[ics[icy]+ny2];
    	point.z=yaray[ics[icy]+ny2];
    	new_point3= new_p_coordinate(point,trans_ps);
    	xpoint[iz*5+2].x = new_point3.x;
    	xpoint[iz*5+2].y = new_point3.y;

    	test=surface_check(new_point1.x,new_point1.y,
      	new_point2.x,new_point2.y,
      	new_point3.x,new_point3.y);
    	if(test>=0.)
    	{
     		point.x=xtem[mx1];point.y=xaray[ics[icy]+ny2];
     		point.z=yaray[ics[icy]+ny2];
     		new_point = new_p_coordinate(point,trans_ps);
     		xpoint[iz*5+3].x = new_point.x;
     		xpoint[iz*5+3].y = new_point.y;
     		xpoint[iz*5+4].x = new_point1.x;
     		xpoint[iz*5+4].y = new_point1.y;
     		zid[iz]= -(new_point1.z+new_point3.z)/2.;
     		iz++;
        }
      } /* end of for(i) */
   } /* end of for(icy) */

   indexx(iz,zid,order);
   for (i=0; i<iz; i++)
   {
      ip=order[i+1]-1;
      fprintf(fp,"%6.2f %6.2f %6.2f %6.2f\n %6.2f %6.2f %6.2f %6.2f Sf\n",
        xpoint[ip*5].x,xpoint[ip*5].y,xpoint[ip*5+1].x,
        xpoint[ip*5+1].y,xpoint[ip*5+2].x,xpoint[ip*5+2].y,
        xpoint[ip*5+3].x,xpoint[ip*5+3].y);
   }
  }  /* end of if(iflg) */

  /* plot curve in 3D space */
  if (*iflg==4)
  {
    fprintf(fp, "0.5 setlinewidth\n");
    j = (theta*phi>0) ? *max_p-1 : 1;
    while(j > 0 && j < *max_p) {
        ny2=0;
        test = phi + 1.570796326795;
        if(test>3.14159265359) test -= 6.28318539718;
        id = (theta*test>0) ? 50 : 1;
        while(id > 0 && id < 51) {

            mx2=(id*(idx2[j-1]-idx1[j-1])+1)/50;
            ny1=((id-1)*(idx2[j]-idx1[j])+1)/50;

            ip=0;
            fprintf(fp, "newpath\n");
            ix=((id-1)*(idx2[j-1]-idx1[j-1])+1)/50;
            while(ix<=mx2){
                point.x = xaray[idx1[j-1]-1+ix+(j-1)*1000];
                point.z = yaray[idx1[j-1]-1+ix+(j-1)*1000];
                point.y = ztem[j-1];
                new_point = new_p_coordinate(point,trans_ps);
                if(ip == 0) {
                    ip++;
                    fprintf(fp, "%6.2f %6.2f moveto\n",
                        new_point.x, new_point.y);
                } else
                    fprintf(fp, "%6.2f %6.2f Lt\n",
                        new_point.x, new_point.y);
                ix++;
            }

            iz=(id*(idx2[j]-idx1[j])+1)/50;
            while(iz>=ny1){
                point.x = xaray[idx1[j]-1+iz+j*1000];
                point.z = yaray[idx1[j]-1+iz+j*1000];
                point.y = ztem[j];
                new_point = new_p_coordinate(point,trans_ps);
                fprintf(fp, "%6.2f %6.2f Lt\n", new_point.x, new_point.y);
                iz--;
            }
            fprintf(fp, "closepath\ngsave\n1 Sg fill Gr S\n");

            (theta*test>0) ? id-- : id++;
        }
        theta*phi>0 ? j-- : j++;
    }
    fprintf(fp, "1 setlinewidth\n");
/*
     for (i=0; i<*max_p; i++)
     {
         mtemp=idx2[i]-idx1[1]+1;
         point.x=xaray[idx1[i]-1+i*1000]; 
         point.z=yaray[idx1[i]-1+i*1000]; 
         point.y=ztem[i];
         new_point=new_p_coordinate(point,trans_ps);
         fprintf(fp," %8.2f %8.2f M\n",new_point.x,new_point.y);
         for (j=1 ; j<mtemp; j++) {
            point.x=xaray[idx1[i]-1+j+i*1000]; 
            point.z=yaray[idx1[i]-1+j+i*1000]; 
            point.y=ztem[i];
            new_point=new_p_coordinate(point,trans_ps);
            fprintf(fp," %8.2f %8.2f Lt\n",new_point.x,new_point.y);
         }
         fprintf(fp,"S\n");
     }
*/
  }
  else
  {
     mtemp=idx2[0]-idx1[0]+1;
     point.x=xtem[idx1[0]-1];
     point.y=ytem[idx1[0]-1];
     point.z=ztem[idx1[0]-1];
     new_point=new_p_coordinate(point,trans_ps);
     fprintf(fp," %8.2f %8.2f M\n",new_point.x,new_point.y);

     for (j=1 ; j<mtemp; j++) {
        point.x=xtem[idx1[0]-1+j];
        point.y=ytem[idx1[0]-1+j];
        point.z=ztem[idx1[0]-1+j];
      new_point=new_p_coordinate(point,trans_ps);
      fprintf(fp," %8.2f %8.2f Lt\n",new_point.x,new_point.y);
     }
     fprintf(fp,"S\n");
  }
}

/***********************************************************
 * end of page          
 */
void ps_3dpage_end_()
{
  fprintf(fp,"restore\n");
  fprintf(fp,"showpage\n");
}

/***********************************************************
 * close PostScript file
 */
void ps3_end_()
{
  int id, ipp, ic;
  char unix_command[80], printer_name[50];

  fprintf(fp,"%%%%Trailer\n");
  fclose(fp);
  /* following code is for printing postscript file
     via system "lpr" command. before using it,
     test "lpr -Pprintername filename" thoroughly
     in unix shell environment */

  printf("\t\tInput 1: Print PostScript File(3D) Now\n");
  printf("\t\t      2: Do Not Print\n");
  scanf("%d", &id);
  if(id==1) {
     printf("\t\tInput 1: Use Printer imagps\n");
     printf("\t\t      2: Use Printer amssps1\n");
     printf("\t\t      3: Use Printer amssps2\n");
     printf("\t\t      4: Select a Printer by Name\n");
     scanf("%d", &ipp);
     switch(ipp)
     {
        case 1:
          printf("\tInput How Many Copies\n");
          scanf("%d", &ic);
          sprintf(unix_command,"lpr -Pimagps -#%d %s",
                ic, fn3d);
          system(unix_command);
          break;
        case 2:
          printf("\tInput How Many Copies\n");
          scanf("%d", &ic);
          sprintf(unix_command,"lpr -Pamssps1 -#%d %s",
                ic, fn3d);
          system(unix_command);
          break;
        case 3:
          printf("\tInput How Many Copies\n");
          scanf("%d", &ic);
          sprintf(unix_command,"lpr -Pamssps2 -#%d %s",
                ic, fn3d);
          system(unix_command);
          break;
        default:
          printf("\tInput Printer Name\n");
          scanf("%s", printer_name);
          printf("\tInput How Many Copies\n");
          scanf("%d", &ic);
          sprintf(unix_command,"lpr -P%s -#%d %s",
                printer_name, ic, fn3d);
          system(unix_command);
          break;
     }
  }
}                 /** ps3d.c **/
